@@ -5,6 +5,7 @@ import hmac
 import logging
 import os
 import os.path as op
+from typing import Tuple
 
 from botocore.exceptions import ClientError
 from copy import deepcopy
@@ -12,9 +13,9 @@ from datetime import datetime, timedelta
 from gzip import GzipFile
 from io import BytesIO
 from os import makedirs, getenv
-from shutil import rmtree
+from shutil import rmtree, copyfileobj
 from tempfile import mkdtemp
-from urllib.parse import urlparse
+from urllib.parse import urlparse, parse_qs
 
 logger = logging.getLogger(__name__)
 
@@ -29,22 +30,28 @@ class s3(object):
 
     @classmethod
     def urlparse(cls, url):
-        """ Split S3 URL into bucket, key, filename """
+        """ Split S3 URL into bucket, key, filename, query_params """
         _url = deepcopy(url)
-        if url[0:5] == 'https':
+
+        if url.startswith("https"):
             _url = cls.https_to_s3(url)
-        if _url[0:5] != 's3://':
-            raise Exception('Invalid S3 url %s' % _url)
 
-        url_obj = _url.replace('s3://', '').split('/')
+        if not url.startswith("s3://"):
+            raise Exception(f"Invalid S3 url {_url}")
 
-        # remove empty items
-        url_obj = list(filter(lambda x: x, url_obj))
+        parsed = urlparse(_url)
+        query_params = parse_qs(parsed.query or "")
+
+        # now fix the "list-of-1" to be a straight string
+        for key in query_params:
+            if len(query_params[key]) == 1:
+                query_params[key] = query_params[key][0]
 
         return {
-            'bucket': url_obj[0],
-            'key': '/'.join(url_obj[1:]),
-            'filename': url_obj[-1] if len(url_obj) > 1 else ''
+            "bucket": parsed.netloc,
+            "key": parsed.path.removeprefix("/"),
+            "filename": os.path.basename(parsed.path),
+            "parameters": query_params,
         }
 
     @classmethod
@@ -69,13 +76,26 @@ class s3(object):
         """ Check if this URL exists on S3 """
         parts = self.urlparse(url)
         try:
-            self.s3.head_object(Bucket=parts['bucket'], Key=parts['key'])
+            kwargs = {}
+            if self.requester_pays:
+                kwargs["RequestPayer"] = "requester"
+
+            self.s3.head_object(Bucket=parts['bucket'],
+                                Key=parts['key'],
+                                **kwargs)
             return True
         except ClientError as exc:
             if exc.response['Error']['Code'] != '404':
                 raise
             return False
 
+    def get_bucket_region(self, bucket_name):
+        region = self.s3.get_bucket_location(
+            Bucket=bucket_name)['LocationConstraint']
+        # US Standard region buckets will have a null location
+        # https://github.com/aws/aws-cli/issues/3864
+        return region if region else 'us-east-1'
+
     def upload(self, filename, url, public=False, extra={}, http_url=False):
         """ Upload object to S3 uri (bucket + prefix), keeping same base filename """
         logger.debug("Uploading %s to %s" % (filename, url))
@@ -89,9 +109,8 @@ class s3(object):
                                    parts['key'],
                                    ExtraArgs=extra)
         if http_url:
-            region = self.s3.get_bucket_location(
-                Bucket=parts['bucket'])['LocationConstraint']
-            return self.s3_to_https(url_out, region)
+            return self.s3_to_https(url_out,
+                                    self.get_bucket_region(parts['bucket']))
         else:
             return url_out
 
@@ -110,40 +129,126 @@ class s3(object):
         finally:
             rmtree(tmpdir)
 
-    def get_object(self, bucket, key):
+    def get_object(self, bucket, key, extra_args={}):
         """ Get an S3 object """
+        extra_args = deepcopy(extra_args or {})
+
         if self.requester_pays:
-            response = self.s3.get_object(Bucket=bucket,
-                                          Key=key,
-                                          RequestPayer='requester')
-        else:
-            response = self.s3.get_object(Bucket=bucket, Key=key)
+            extra_args["RequestPayer"] = "requester"
+
+        return self.s3.get_object(Bucket=bucket, Key=key, **extra_args)
+
+    def get_object_metadata(self, uri, **kwargs):
+        """
+        Get object metadata attributes.
+
+        Additional keyword arguments are passed to the underlying client.
+
+        :param uri: URI of the object
+        """
+        s3_uri = self.urlparse(uri)
+        logger.debug("Retrieving object attributes for %s", uri)
+
+        extra_args = deepcopy(
+            s3_uri["parameters"]) if "parameters" in s3_uri else {}
+        if kwargs:
+            for k in kwargs:
+                extra_args[k] = kwargs[k]
+        if self.requester_pays:
+            extra_args["RequestPayer"] = "requester"
+
+        response = self.s3.get_object_attributes(
+            Bucket=s3_uri["bucket"],
+            Key=s3_uri["key"],
+            ObjectAttributes=["ETag", "Checksum", "ObjectSize", "ObjectParts"],
+            **extra_args)
+
+        response.pop("ResponseMetadata", None)
+
         return response
 
-    def download(self, uri, path=''):
+    def download(self, uri, path='', **kwargs):
         """
         Download object from S3
+
+        Additional keyword parameters will be passed to the underlying client.
+
         :param uri: URI of object to download
         :param path: Output path
         """
         s3_uri = self.urlparse(uri)
         fout = op.join(path, s3_uri['filename'])
         logger.debug("Downloading %s as %s" % (uri, fout))
-        if path != '':
+
+        if path != "":
             makedirs(path, exist_ok=True)
-        extra_args = None
+
+        extra_args = deepcopy(kwargs) if kwargs else {}
         if self.requester_pays:
-            extra_args = {'RequestPayer': 'requester'}
-        self.s3.download_file(s3_uri['bucket'],
-                              s3_uri['key'],
+            extra_args["RequestPayer"] = "requester"
+
+        if s3_uri["parameters"]:
+            for key in s3_uri["parameters"]:
+                extra_args[key] = s3_uri["parameters"][key]
+
+        self.s3.download_file(s3_uri["bucket"],
+                              s3_uri["key"],
                               fout,
                               ExtraArgs=extra_args)
         return fout
 
+    def download_with_metadata(self,
+                               uri,
+                               path='',
+                               extra_args={}) -> Tuple[str, dict]:
+        """Download object from S3.
+
+        :param uri: URI of object to download
+        :param path: Output path
+        :param extra_args: Extra arguments passed to the s3 client.
+        """
+        s3_uri = self.urlparse(uri)
+        fout = os.path.join(path, s3_uri["filename"])
+
+        if path != "":
+            os.makedirs(path, exist_ok=True)
+
+        extra_args = deepcopy(extra_args or {})
+        if self.requester_pays:
+            extra_args["RequestPayer"] = "requester"
+
+        if s3_uri["parameters"]:
+            for key in s3_uri["parameters"]:
+                extra_args[key] = s3_uri["parameters"][key]
+
+        result = self.get_object(s3_uri["bucket"],
+                                 s3_uri["key"],
+                                 extra_args=extra_args)
+
+        # save file
+        with open(fout, "wb") as f:
+            copyfileobj(result["Body"], f)
+
+        # populate metadata dictionary
+        meta = deepcopy(result["Metadata"]) if "Metadata" in result else {}
+        for key in [
+                "LastModified", "Expiration", "ETag", "VersionId",
+                "ContentEncoding", "ContentLanguage", "ContentType", "Expires",
+                "ChecksumCRC32", "ChecksumCRC32C", "ChecksumSHA1",
+                "ChecksumSHA256"
+        ]:
+            if key in result:
+                meta[key] = result[key]
+
+        return (fout, meta)
+
     def read(self, url):
         """ Read object from s3 """
         parts = self.urlparse(url)
-        response = self.get_object(parts['bucket'], parts['key'])
+        kwargs = {}
+        if self.requester_pays:
+            kwargs["RequestPayer"] = "requester"
+        response = self.get_object(parts['bucket'], parts['key'], **kwargs)
         body = response['Body'].read()
         if op.splitext(parts['key'])[1] == '.gz':
             body = GzipFile(None, 'rb', fileobj=BytesIO(body)).read()
@@ -156,8 +261,8 @@ class s3(object):
     def delete(self, url):
         """ Remove object from S3 """
         parts = self.urlparse(url)
-        response = self.s3.delete_object(Bucket=parts['Bucket'],
-                                         Key=parts['Key'])
+        response = self.s3.delete_object(Bucket=parts['bucket'],
+                                         Key=parts['key'])
         return response
 
     # function derived from https://alexwlchan.net/2018/01/listing-s3-keys-redux/
@@ -175,6 +280,9 @@ class s3(object):
         if isinstance(parts['key'], str):
             kwargs['Prefix'] = parts['key']
 
+        if self.requester_pays:
+            kwargs["RequestPayer"] = "requester"
+
         while True:
             # The S3 API response is a large blob of metadata.
             # 'Contents' contains information about the listed objects.
@@ -204,6 +312,8 @@ class s3(object):
                             suffix=None,
                             start_date=None,
                             end_date=None,
+                            is_latest=None,
+                            key_contains=None,
                             datetime_key='LastModifiedDate'):
         logger.debug('Reading inventory file %s' % (fname))
 
@@ -231,7 +341,24 @@ class s3(object):
                                    "%Y-%m-%dT%H:%M:%S.%fZ").date()
             return True if dt < end_date else False
 
+        def islatest(info):
+            latest = info.get("IsLatest")
+            if latest:
+                return latest not in ("false", False)
+            return True
+
+        def fcontains(info):
+            for part in key_contains:
+                if part not in info["Key"]:
+                    return False
+            return True
+
         inv = filter(fvalid, inv)
+
+        if is_latest is not None:
+            inv - filter(islatest, inv)
+        if key_contains:
+            inv = filter(fcontains, inv)
         if prefix:
             inv = filter(fprefix, inv)
         if suffix:
@@ -240,30 +367,34 @@ class s3(object):
             inv = filter(fstartdate, inv)
         if end_date:
             inv = filter(fenddate, inv)
+
         for i in inv:
             yield 's3://%s/%s' % (i['Bucket'], i['Key'])
 
-    def latest_inventory_manifest(self, url):
+    def latest_inventory_manifest(self, url, manifest_age_days=1):
         """ Get latest inventory manifest file """
         parts = self.urlparse(url)
         # get latest manifest file
         today = datetime.now()
         manifest_url = None
-        for dt in [today, today - timedelta(1)]:
+        for dt in [today - timedelta(x) for x in range(manifest_age_days)]:
             _key = op.join(parts['key'], dt.strftime('%Y-%m-%d'))
             _url = 's3://%s/%s' % (parts['bucket'], _key)
             manifests = [k for k in self.find(_url, suffix='manifest.json')]
             if len(manifests) == 1:
                 manifest_url = manifests[0]
                 break
+
         if manifest_url:
             return self.read_json(manifest_url)
         else:
             return None
 
-    def latest_inventory_files(self, url):
+    def latest_inventory_files(self, url, manifest=None):
+        if not manifest:
+            manifest = self.latest_inventory_manifest(url)
+
         bucket = self.urlparse(url)['bucket']
-        manifest = self.latest_inventory_manifest(url)
         if manifest:
             files = manifest.get('files', [])
             numfiles = len(files)
@@ -276,7 +407,9 @@ class s3(object):
 
     def latest_inventory(self, url, **kwargs):
         """ Return generator function for objects in Bucket with suffix (all files if suffix=None) """
-        manifest = self.latest_inventory_manifest(url)
+        manifest_age_days = kwargs.pop("manifest_age_days", 1)
+        manifest = self.latest_inventory_manifest(url, manifest_age_days)
+
         # read through latest manifest looking for matches
         if manifest:
             # get file schema
@@ -284,7 +417,8 @@ class s3(object):
                 str(key).strip() for key in manifest['fileSchema'].split(',')
             ]
 
-            for i, url in enumerate(self.latest_inventory_files(url)):
+            for i, url in enumerate(self.latest_inventory_files(url,
+                                                                manifest)):
                 logger.info('Reading inventory file %s' % (i + 1))
                 results = self.read_inventory_file(url, keys, **kwargs)
                 yield from results
